/*
** Module   : character.java
** Abstract : Progress 4GL compatible character object
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050517   @21295 Created initial version with full support
**                           for all Progress character processing
**                           including the concatination, begins, matches
**                           operators (but not CONTAINS which is not
**                           needed at this time), all builtin functions,
**                           case-sensitivity and the unknown value.
** 002 GES 20050606   @21424 Minor additions for function support.
** 003 GES 20050621   @21440 Added an assign() method to modify this
**                           instance's state based on the result of
**                           an expression of compatible type.
** 004 GES 20050627   @21554 Added replace/overlay methods to provide the
**                           substring/overlay assignment type language
**                           statements.
** 005 GES 20050628   @21565 Fixed typo in method name and added unknown
**                           value processing to toStringMessage().
** 006 SVG 20050720   @21768 Cloning support for BaseDataType added.
** 007 NVS 20050728   @21908 Added implementation of PROGRESS builtin fun-
**                           ction CAN-DO as public static canDo, taking
**                           either one or two parameters. One parameter
**                           version is just a catcher for untranslated
**                           CAN-DO calls, throwing an exception. Two par-
**                           ameter version is a real code which is typic-
**                           ally used as a utility function.
** 008 GES 20050728   @21920 Added support for the _cbit() built-in 
**                           function (see the testBitAt() method). Also
**                           changed the name of canDo to matchesList
**                           which is more generic and less like a
**                           Progress security-oriented function.
** 009 GES 20050815   @22118 Added Undoable interface support.
** 010 GES 20050825   @22253 Fixed toStringExport() to double up double
**                           quotes appearing inside the string.
** 011 GES 20050826   @22270 Implement case-sensitivity in output from
**                           specific functions and operators depending
**                           on the case-sensitivity of character operands
**                           or parameters.
** 012 GES 20050828   @22279 Added support for temporarily setting
**                           case-sensitivity with an automatic reset at
**                           the end of the scope in which the setting
**                           was made.
** 013 GES 20050829   @22284 Split assignment of configuration and state
**                           from normal assign(). deepAssign() is now 
**                           provided to set both configuration and state
**                           at the same time.
** 014 GES 20050829   @22371 Added error condition support.
** 015 GES 20050907   @22594 Added "raw" type support to substring().
** 016 ECF 20051004   @22958 Implement hashCode() method. Required to be
**                           consistent with BaseDataType's implementation
**                           of equals() method.
** 017 GES 20051024   @23122 Added convertToSQLLike() method.
** 018 ECF 20051104   @23237 Added iterate() method to SensitivityResetter
**                           inner class. No-op implementation; required
**                           to satisfy expanded Finalizable interface.
** 019 GES 20060112   @23892 Added additional valueOf signatures for the
**                           literal forms of the string(), fill(),
**                           entry(), indexOf() and lastIndexOf()
**                           functions.
** 020 GES 20060114   @23903 Fixed substring (2nd parm was used wrong).
** 021 GES 20060206   @24297 Made compareTo() safe to process any object
**                           type (avoids ClassCastException).
** 022 GES 20060208   @24387 Expose a format string expansion helper as
**                           public and added a default format string
**                           getter.
** 023 GES 20060303   @24882 Don't escape the single quote char in Java 
**                           strings.
** 024 GES 20060312   @24985 New variants for overlay and replace to
**                           handle all possible combos with primitives.
** 025 ECF 20060322   @25203 Honor case sensitivity in indexOf() methods.
**                           Added new indexOf variant to which all others
**                           delegate.  Performs scan with proper case
**                           sensitivity setting.
** 026 GES 20060403   @25318 New variants for lookup().
** 027 GES 20060410   @25457 New variants for substring().
** 028 GES 20060417   @25552 Added default value generation.
** 029 GES 20060420   @25620 Added new signatures for right/leftTrim.
** 030 GES 20060422   @25650 Changed assignment style syntax backing
**                           methods to return "this" such that the result
**                           can be assigned back in the case of a 
**                           database field. Added entry statement 
**                           support.
** 031 GES 20060427   @25762 New signatures for trim().
** 032 GES 20060427   @25767 New signature for length().
** 033 GES 20060508   @26023 Fix for Progress pattern to regular 
**                           expression conversion routine
**                           (convertToRegEx).
** 034 ECF 20060510   @26062 Introduced a runtime variant of static method
**                           convertToSQLLike(). This variant accepts a
**                           character parameter.
** 035 GES 20060517   @26192 Match new Finalizable interface requirement.
** 036 GES 20060518   @26280 Safety code for negative start indexes in
**                           substring.
** 037 GES 20060525   @26530 Fix for asc() which needed a 0-based index.
** 038 ECF 20060613   @27243 Fix for asc(). Replaced implementation to
**                           convert the low order character byte to an
**                           unsigned integer.
** 039 GES 20060724   @28158 Added a forced assign() abstract method.
** 040 GES 20060807   @28471 Moved to externalizable interface to optimize
**                           network performance.
** 041 ECF 20060812   @28586 Added methods needed for database server-side
**                           function support.
** 042 GES 20060904   @29214 Added helpers to space-ify strings with an
**                           embedded null character or null char escape
**                           sequence.
** 043 GES 20060905   @29222 Added embedded null processing for the
**                           toString() and concat() variants. Fixed
**                           null byte processing for asc().
** 044 GES 20060905   @29228 Added safeDuplicate().
** 045 GES 20060907   @29282 Fixed substring processing in "raw" mode.
** 046 GES 20060908   @29346 Fixed lookup() to return 0 instead of -1 on
**                           failure.
** 047 ECF 20060908   @29384 Made constructors which accept a J2SE String
**                           null-safe. A null string argument results in
**                           the instance being set to unknown value.
** 048 GES 20061109   @31050 Added formatLength().
** 049 ECF 20070306   @32284 Added case-sensitive variants of equals and
**                           hashCode. These are equalsCaseSensitively and
**                           caseSensitiveHashCode. Also modified existing
**                           hashCode method to be more consistent with
**                           equals. The previous implementation was not
**                           consistent with equals for case-insensitive
**                           character instances.
** 050 GES 20070321   @32504 Fix for some forms of lastIndexOf() which
**                           were starting the backwards search from the
**                           beginning of the source string instead of the 
**                           end as it should have been. This affected
**                           all forms of the method which only have 2
**                           parameters. Also fixed replaceAll() which
**                           would fail to replace two back-to-back
**                           instances of the same target string (the
**                           first would be replaced and the first char
**                           of the second match would be dropped).
** 051 GES 20070926   @35243 Provide external control over 4GL compatible
**                           escape characters when processing non-runtime
**                           strings in progressSpacifyNull(). On Windows
**                           that code should not treat the backslash as
**                           an escape character.
** 052 GES 20080312   @37442 Fixes and documentation added to format
**                           string processing. Fill characters and
**                           padding were not always being handled
**                           properly. Data characters were not being
**                           copied into the right locations in the output
**                           string when non-trailing fill characters were
**                           present.
** 053 GES 20080414   @38012 Fixed chr/asc to support all non-DBCS
**                           charsets. Added fromExportString().
** 054 ECF 20080804   @39286 Added 2 new flavors of numEntries(). Needed
**                           to handle combinations of character and
**                           String parameters.
** 055 GES 20080909   @39766 Moved concat() to varargs. This makes it
**                           possible to merge multiple concat operator
**                           usage into a single method call, which is
**                           significantly more efficient and cleaner.
** 056 GES 20080911   @39788 Added assign(String) signature.
** 057 GES 20081002   @40033 Modified method names to shorten calling
**                           code.
** 058 SVL 20081224   @40985 valueOf(..) returns unknown character value
**                           if converted data is an unknown value.
** 059 ECF 20090212   @41296 Made class final.
** 060 GES 20090422   @41905 Converted to standard string formatting.
** 061 GES 20090424   @41968 Import change.
** 062 GES 20090706   @43106 Added signatures for overlay().
** 063 ECF 20090814   @43621 Fixed various methods which use delimiters. Added
**                           replaceEntry(double, character, String) variant.
** 064 ECF 20090816   @43659 Rolled back #059 (@41296).
** 065 GES 20110303          Moved maximum/minimum to varargs.
** 066 GES 20110422          Fixed case-sensitivity inheritance in the return
**                           value of concat().
** 067 SVL 20110622          Modified convertToSQLLike, added
**                           convertToSQLBegins, postprocessStringLiteral,
**                           toUpperCase(String, boolean) and
**                           substring(String, int).
** 068 GES 20110801          Modified substring() signatures to use NumberType.
** 069 CS  20121119          Implement the quoter built-in function.
** 070 CS  20121231          Added String overloads methods for quoter 
**                           built-in function
** 071 EVL 20130107          Changing progressToJavaString() adding the version
**                           of the method that makes special processing for
**                           '\' path related strings for Windows-target case.
** 072 CA  20130123          Added c'tor which builds a new instance from a
**                           BaseDataType value, if possible. This should be
**                           used only when constructing an instance using the
**                           value returned by a DYNAMIC-FUNCTION call; 4GL
**                           does some special casting in this case, which is
**                           not implemented yet.
** 073 OM  20121023          Added support for TO-ROWID(STRING(row_id)).
** 074 AIL 20130202          Moved static methods common to character and 
**                           longchar to TextOps class.
** 075 OM  20130212          Added support CONTAINS clause.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import com.goldencode.util.*;

/**
 * A class that represents a Progress 4GL compatible character object whose
 * data is mutable.  All Progress language features related to character
 * types are supported including all functionality that can be accessed via
 * the character concatination operator and builtin Progress string
 * functions.  Normal character comparison operators are implemented in
 * the {@link CompareOps} class. Operator and functions common to character 
 * and longchar types are implemented as statics in the {@link TextOps} 
 * class.
 * <p>
 * In addition to the implementation of the base string processing functions
 * and operators, this class hides the representation of the 
 * <code>unknown value</code> and properly maintains the processing of
 * <code>unknown value</code>. 
 * <p>
 * Format string related processing is handled in this class, see method
 * {@link #toString(String)}.
 * <p>
 * The following is the mapping of Progress language features to the
 * corresponding feature in this class:
 * <p>
 * <pre>
 * contains         operator                       not supported at this time
 * _cbit            function                       {@link #testBitAt}
 * chr              function                       {@link #chr}
 * asc              function                       {@link #asc}
 * string           function                       {@link #valueOf}
 * fill             function                       {@link #fill}
 * quoter           function                       {@link #quoter} 
 * </pre> 
 * <p>
 * <b>Please note that at this time, no conversion of Progress strings to
 * Java strings is done.  All strings passed are expected to be Java
 * strings.  This behavior may need changing in the case where a Progress
 * string is built at runtime but needs to be used as a Java string.</b> In
 * this case the {@link #progressToJavaString} method should be used.
 *
 * @author    GES
 */
public class character
extends Text
{
   /** The default format string. */
   private static final String defaultFormat = "x(8)";
   
   /**
    * Default constructor, creates an instance that represents the unknown
    * value.
    */
   public character()
   {
      super();
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a DYNAMIC-FUNCTION call to the expected type (i.e
    * when passing that value as a parameter to another function or procedure
    * call).  In such cases, 4GL does some automatic conversion, which is not
    * implemented yet.
    * <p>
    * This version of the c'tor always sets the value to unknown.
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public character(BaseDataType value)
   {
      super(value);
   }
   
   /**
    * Create a new instance, by copying the state of the passed {@link Text}.
    * 
    * @param    value
    *           The value which needs to be copied.
    */
   public character(Text value)
   {
      super(value, false);
   }

   /**
    * Create a new instance, by copying the state of the passed {@link Text}.
    * If the <code>output</code> parameter is set to <code>true</code>, it 
    * means that the referenced value is saved, as the passed instance will
    * be passed to a function call, and the state might need to be updated
    * on function exit. 
    * 
    * @param    value
    *           The value which needs to be copied.
    * @param    output
    *           Flag indicating if this instance is passed to as a reference
    *           for an output parameter.
    */
   public character(Text value, boolean output)
   {
      super(value, output);
   }
   
   /**
    * Constructs an instance that has the <code>String</code> value passed
    * in and which <b>does not</b> represent the <code>unknown value</code>,
    * unless the <code>value</code> argument is <code>null</code>.
    *
    * @param    value
    *           The character value of the resulting instance.
    */
   public character(String value)
   {
      super(value);
   }
   
   /**
    * Constructs an instance that has the <code>String</code> value passed
    * in and which <b>does not</b> represent the <code>unknown value</code>
    * unless the <code>value</code> argument is <code>null</code>.
    *
    * @param    value
    *           The character value of the resulting instance.
    * @param    caseSens
    *           Specifies whether this instance should use case sensitive
    *           comparisons.
    */
   public character(String value, boolean caseSens)
   {
      super(value, caseSens);
   }
   
   /**
    * Constructs an instance that has a <code>char</code> value passed
    * in and which <b>does not</b> represent the <code>unknown value</code>. 
    *
    * @param    ch
    *           The character value of the resulting instance.
    */
   public character(char ch)
   {
      super(ch);
   }
   
   /**
    * Constructs an instance that has the <code>String</code> value and
    * <code>unknown value</code> that exactly matches that of the passed-in 
    * instance.
    *
    * @param    value
    *           The character and unknown value of the resulting instance.
    */
   public character(character value)
   {
      super(value);
   }

   /**
    * Returns the numeric value of a character expression which must be
    * only 1 character in length.  This method will probably generate
    * incorrect results in a DBCS environment since it will use UNICODE
    * as the basis for conversion.  If this returned value is different than
    * would be returned under Progress AND if that Progress-specific value 
    * is used in comparison with this returned value, then a difference in
    * logic will be seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    *
    * @param    c
    *           The character expression to convert (must be 1 character in
    *           length).
    *
    * @return    The numeric value of that character which should match up
    *            with the Progress equivalents in the ASCII range.
    */
   public static integer asc(character c)
   {
      if (c.isUnknown())
      {
         return new integer();
      }
      
      return asc(c.getValue());
   }
      
   /**
    * Returns the numeric value of a string which must be only 1 character
    * in length.  This method will probably generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for conversion.
    * If this returned value is different than would be returned under
    * Progress AND if that Progress-specific value is used in comparison with
    * this returned value, then a difference in logic will be seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    *
    * @param    str
    *           The string to convert (must be 1 character in length).
    *
    * @return   The numeric value of that character which should match up
    *           with the Progress equivalents in the ASCII range.
    */
   public static integer asc(String str)
   {
      int i = -1;
      
      if (str.length() == 1)
      {
         i = Utils.toInt(str.charAt(0), override);
      }
      
      return new integer(i);
   }
      
   /**
    * Returns the character value of a numeric expression which must be
    * between 1 and 65534 (inclusive).  This method will probably generate
    * incorrect results in a DBCS environment since it will use UNICODE
    * as the basis for conversion.  If this returned value is different than
    * would be returned under Progress AND if that Progress-specific value 
    * is used in comparison with this returned value, then a difference in
    * logic will be seen.  The result should match up with the Progress
    * equivalents in the ASCII range.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    *
    * @param    i
    *           The integer to convert.
    *
    * @return   The associated character or a character representing the
    *           empty string (0 length) when the input is out of range or
    *           <code>unknown value</code> if the input is
    *           <code>unknown value</code>.
    */
   public static character chr(integer i)
   {
      if (i.isUnknown())
      {
         return new character();
      }
      
      return chr(i.intValue());
   }
      
   /**
    * Returns the character value of a numeric expression which must be
    * between 1 and 65534 (inclusive).  This method will probably generate
    * incorrect results in a DBCS environment since it will use UNICODE
    * as the basis for conversion.  If this returned value is different than
    * would be returned under Progress AND if that Progress-specific value 
    * is used in comparison with this returned value, then a difference in
    * logic will be seen.  The result should match up with the Progress
    * equivalents in the ASCII range.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    *
    * @param    i
    *           The integer to convert.
    *
    * @return   The associated character or a character representing the
    *           empty string (0 length) when the input is out of range.
    */
   public static character chr(int i)
   {
      if (i < 1 || i > 65534)
      {
         return new character("");
      }
      
      char ch = Utils.toChar(i, override);
      
      return (ch == 0xFFFF) ? new character("") : new character(ch);
   }
      
   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    c
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(character c, NumberType num)
   {
      if (c.isUnknown() || num.isUnknown())
      {
         return new character();
      }
      
      return fill(c.getValue(), num.intValue());
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    c
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(character c, int num)
   {
      if (c.isUnknown())
      {
         return new character();
      }
      
      return fill(c.getValue(), num);
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    c
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(character c, double num)
   {
      if (c.isUnknown())
      {
         return new character();
      }
      
      return fill(c.getValue(), (int) num);
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    str
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(String str, NumberType num)
   {
      if (num.isUnknown())
      {
         return new character();
      }
      
      return fill(str, num.intValue());
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    str
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(String str, double num)
   {
      return fill(str, (int) num);
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    str
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(String str, int num)
   {
      if (num < 1)
         return new character("");
      
      StringBuffer sb = new StringBuffer();
      
      for (int i = 0; i < num; i++)
      {
         sb.append(str);
      }
      
      return new character(sb.toString());
   }

   /**
    * General quoter function implementation without additional parameters.
    * <p>
    * This puts the <code>character</code> value of the variable between
    * double quotes if the value is not <code>unknown</code>  or replace it 
    * with '?' if the value is <code>unknown</code>.
    * 
    * @param     var
    *            The <code>BaseDataType</code> variable to operate upon.
    *
    * @return    The resulting quoted character value of the function.
    */
   public static character quoter(BaseDataType var)
   {
      return quoterWorker(var, null, null);
   }
   
   /**
    * General quoter function implementation without additional parameters.
    * <p>
    * This puts the <code>character</code> value of the variable between
    * double quotes if the value is not <code>null</code>  or replace it 
    * with '?' if the value is <code>null</code>.
    * 
    * @param     var
    *            The <code>String</code> variable to operate upon.
    *
    * @return    The resulting quoted character value of the function.
    */
   public static character quoter(String var)
   {
      return quoterWorker(new character(var), null, null);
   }
   
   /**
    * No parameter quoter function implementation.
    * This will be called when trying to quote an unknown literal.
    * 
    * @return    The default unknown character ?.
    */
   public static character quoter()
   {
      return new character("?");
   }
   
   /**
    * Quoter function call for 2 parameters.
    * <p>
    * This puts the <code>character</code> value of the variable between
    * double quotes if the value is not <code>unknown</code>  or replace it 
    * with '?' if the value is <code>unknown</code>.
    * <p>
    * Additionally a parameter representing a custom  quote character 
    * replacement is provided. If not <code>null</code>, then the first
    * character of the given text will be used instead of the &quot; 
    * character.
    * <p>
    * A custom unknown value can only be given in the 3 parameter call.
    *  
    * @param    var 
    *           The <code>BaseDataType</code> variable to operate upon
    * @param    customQuote
    *           The <code>String</code> variable
    *           that represents the custom quoter.Custom <code>unknown</code> 
    *           can only be given in 3 parameter call.
    *
    * @return   The resulting quoted character value of the function.
    */
   public static character quoter(BaseDataType var, String customQuote)
   {
      return quoterWorker(var, customQuote, null);
   }
   
   /**
    * Quoter function call for 2 parameters.
    * <p>
    * This puts the <code>String</code> value of the variable between
    * double quotes if the value is not <code>null</code>  or replace it 
    * with '?' if the value is <code>null</code>.
    * <p>
    * Additionally a parameter representing a custom  quote character 
    * replacement is provided. If not <code>null</code>, then the first
    * character of the given text will be used instead of the &quot; 
    * character.
    * <p>
    * A custom unknown value can only be given in the 3 parameter call.
    *  
    * @param    var 
    *           The <code>String</code> variable to operate upon
    * @param    customQuote
    *           The <code>String</code> variable
    *           that represents the custom quoter.Custom <code>unknown</code> 
    *           can only be given in 3 parameter call.
    *
    * @return   The resulting quoted character value of the function.
    */
   public static character quoter(String var, String customQuote)
   {
      return quoterWorker(new character(var), customQuote, null);
   }
   
   /**
    * Quoter function call for 2 parameters.
    * <p>
    * This puts the <code>character</code> value of the variable between
    * double quotes if the value is not <code>unknown</code>  or replace it 
    * with '?' if the value is <code>unknown</code>.
    * <p>
    * Additionally a parameter representing a custom character 
    * replacement is provided. If not <code>null</code>, then the first
    * character of the given text will be used instead of the &quot; 
    * character.
    * <p>
    * A custom unknown value can only be given in the 3 parameter call.
    *  
    * @param    var 
    *           The <code>BaseDataType</code> variable to operate upon
    * @param    customQuote
    *           The <code>String</code> variable
    *           that represents the custom quoter.Custom <code>unknown</code> 
    *           can only be given in 3 parameter call.
    *
    * @return   The resulting quoted character value of the function.
    */
   public static character quoter(BaseDataType var, character customQuote)
   {
      return quoter(var, customQuote, (String) null);
   }
   
   /**
    * Quoter function call for 2 parameters.
    * <p>
    * This puts the <code>String</code> value of the variable between
    * double quotes if the value is not <code>null</code>  or replace it 
    * with '?' if the value is <code>null</code>.
    * <p>
    * Additionally a parameter representing a custom character 
    * replacement is provided. If not <code>null</code>, then the first
    * character of the given text will be used instead of the &quot; 
    * character.
    * <p>
    * A custom unknown value can only be given in the 3 parameter call.
    *  
    * @param    var 
    *           The <code>String</code> variable to operate upon
    * @param    customQuote
    *           The <code>String</code> variable
    *           that represents the custom quoter.Custom <code>unknown</code> 
    *           can only be given in 3 parameter call.
    *
    * @return   The resulting quoted character value of the function.
    */
   public static character quoter(String var, character customQuote)
   {
      return quoter(new character(var), customQuote, (String) null);
   }
   
   /**
    * Quoter function implementation for 3 parameters.
    * <p>
    * This puts the <code>character</code> value of the variable between
    * double quotes if the value is not <code>unknown</code> or replace it 
    * with '?' if the value is <code>unknown</code>.
    * <p>
    * Additionally, a parameter representing a custom quote character
    * is provided.  If not <code>null</code>, then the first character
    * of the given text will be used instead of the &quot; 
    * character.
    * <p>
    * A custom unknown value can also be provided.  If
    * not <code>null</code>, then the given text will be used instead
    * of the default ? <code>unknown</code> replacement to be used.
    *  
    * 
    * @param    var
    *           The <code>BaseDataType</code> variable to operate upon.
    * @param    customQuote
    *           The custom <code>character</code> quoter variable.
    * @param    customNull
    *           The custom <code>character</code> <code>unknown</code> 
    *           replacement to be used.
    *
    * @return   The resulting quoted <code>character</code> value 
    *           of the function.
    */
   
   public static character quoter(BaseDataType var,
                                  character customQuote,
                                  character customNull)
   {
      return quoterWorker(var,
                          customQuote.isUnknown() ? null 
                                                  : customQuote.getValue(),
                          customNull.isUnknown() ? null 
                                                 : customNull.getValue());
   }
   
   /**
    * Quoter function implementation for 3 parameters.
    * <p>
    * This puts the <code>String</code> value of the variable between
    * double quotes if the value is not <code>null</code> or replace it 
    * with '?' if the value is <code>null</code>.
    * <p>
    * Additionally, a parameter representing a custom quote character
    * is provided.  If not <code>null</code>, then the first character
    * of the given text will be used instead of the &quot; 
    * character.
    * <p>
    * A custom unknown value can also be provided.  If
    * not <code>null</code>, then the given text will be used instead
    * of the default ? <code>unknown</code> replacement to be used.
    *  
    * 
    * @param    var
    *           The <code>String</code> variable to operate upon.
    * @param    customQuote
    *           The custom <code>character</code> quoter variable.
    * @param    customNull
    *           The custom <code>character</code> <code>unknown</code> 
    *           replacement to be used.
    *
    * @return   The resulting quoted <code>character</code> value 
    *           of the function.
    */
   public static character quoter(String var,
                                  character customQuote,
                                  character customNull)
   {
      return quoterWorker(new character(var), 
                          customQuote.isUnknown() ? null 
                                                  : customQuote.getValue(),
                          customNull.isUnknown() ? null 
                                                 : customNull.getValue());
   }
   
   /**
    * Quoter function implementation for 3 parameters.
    * <p>
    * This puts the <code>character</code> value of the variable between
    * double quotes if the value is not <code>unknown</code> or replace it 
    * with '?' if the value is <code>unknown</code>.
    * <p>
    * Additionally, a parameter representing a custom quote character
    * is provided.  If not <code>null</code>, then the first character
    * of the given text will be used instead of the &quot; 
    * character.
    * <p>
    * A custom unknown value can also be provided.  If
    * not <code>null</code>, then the given text will be used instead
    * of the default ? <code>unknown</code> replacement to be used.
    * 
    * @param    var
    *           The <code>BaseDataType</code> variable to operate upon.
    * @param    customQuote
    *           The custom <code>String</code> quoter variable.
    * @param    customNull
    *           The custom <code>character</code> <code>unknown</code> 
    *           replacement to be used.
    *
    * @return   The resulting quoted <code>character</code> value 
    *           of the function.
    */
   public static character quoter(BaseDataType var,
                                  String customQuote,
                                  character customNull)
   {
      return quoterWorker(var, 
                          customQuote, 
                          customNull.isUnknown() ? null 
                                                 : customNull.getValue());
   }
   
   /**
    * Quoter function implementation for 3 parameters.
    * <p>
    * This puts the <code>String</code> value of the variable between
    * double quotes if the value is not <code>null</code> or replace it 
    * with '?' if the value is <code>null</code>.
    * <p>
    * Additionally, a parameter representing a custom quote character
    * is provided.  If not <code>null</code>, then the first character
    * of the given text will be used instead of the &quot; 
    * character.
    * <p>
    * A custom unknown value can also be provided.  If
    * not <code>null</code>, then the given text will be used instead
    * of the default ? <code>unknown</code> replacement to be used.
    * 
    * @param    var
    *           The <code>String</code> variable to operate upon.
    * @param    customQuote
    *           The custom <code>String</code> quoter variable.
    * @param    customNull
    *           The custom <code>character</code> <code>unknown</code> 
    *           replacement to be used.
    *
    * @return   The resulting quoted <code>character</code> value 
    *           of the function.
    */
   public static character quoter(String var,
                                  String customQuote,
                                  character customNull)
   {
      return quoterWorker(new character(var), 
                          customQuote, 
                          customNull.isUnknown() ? null 
                                                 : customNull.getValue());
   }
   
   /**
    * Quoter function implementation for 3 parameters.
    * <p>
    * This puts the <code>character</code> value of the variable between
    * double quotes if the value is not <code>unknown</code> or replace it 
    * with '?' if the value is <code>unknown</code>.
    * <p>
    * Additionally, a parameter representing a custom quote character
    * is provided.  If not <code>null</code>, then the first character
    * of the given text will be used instead of the &quot;
    * character.
    * <p>
    * A custom unknown value can also be provided.  If
    * not <code>null</code>, then the given text will be used instead
    * of the default ? <code>unknown</code> replacement to be used.
    *  
    * @param    var
    *           The <code>BaseDataType</code> variable to operate upon.
    * @param    customQuote
    *           The custom <code>character</code> quoter variable.
    * @param    customNull
    *           The custom <code>String</code> <code>unknown</code> 
    *           replacement to be used.
    *
    * @return   The resulting quoted <code>character</code> value 
    *           of the function.
    */
   public static character quoter(BaseDataType var,
                                  character customQuote,
                                  String customNull)
   {
      return quoterWorker(var,
                          customQuote.isUnknown() ? null 
                                                  : customQuote.getValue(),
                          customNull);
   }
   /**
    * Quoter function implementation for 3 parameters.
    * <p>
    * This puts the <code>String</code> value of the variable between
    * double quotes if the value is not <code>null</code> or replace it 
    * with '?' if the value is <code>null</code>.
    * <p>
    * Additionally, a parameter representing a custom quote character
    * is provided.  If not <code>null</code>, then the first character
    * of the given text will be used instead of the &quot;
    * character.
    * <p>
    * A custom unknown value can also be provided.  If
    * not <code>null</code>, then the given text will be used instead
    * of the default ? <code>unknown</code> replacement to be used.
    *  
    * @param    var
    *           The <code>String</code> variable to operate upon.
    * @param    customQuote
    *           The custom <code>character</code> quoter variable.
    * @param    customNull
    *           The custom <code>String</code> <code>unknown</code> 
    *           replacement to be used.
    *
    * @return   The resulting quoted <code>character</code> value 
    *           of the function.
    */
   public static character quoter(String var,
                                  character customQuote,
                                  String customNull)
   {
      return quoterWorker(new character(var),
                          customQuote.isUnknown() ? null 
                                                  : customQuote.getValue(),
                          customNull);
   }
   
   /**
    * Quoter function implementation for 3 parameters.
    * <p>
    * This puts the <code>character</code> value of the variable between
    * double quotes if the value is not <code>unknown</code> or replace it 
    * with '?' if the value is <code>unknown</code>.
    * <p>
    * Additionally, a parameter representing a custom quote character
    * is provided.  If not <code>null</code>, then the first character
    * of the given text will be used instead of the &quot; 
    * character.
    * <p>
    * A custom unknown value can also be provided.  If
    * not <code>null</code>, then the given text will be used instead
    * of the default ? <code>unknown</code> replacement to be used.
    *  
    * @param    var
    *           The <code>BaseDataType</code> variable to operate upon.
    * @param    customQuote
    *           The custom <code>String</code> quoter variable.
    * @param    customNull
    *           The custom <code>String</code> <code>unknown</code> 
    *           replacement to be used.
    *
    * @return   The resulting quoted <code>character</code> value 
    *           of the function.
    */
   public static character quoter(BaseDataType var,
                                  String customQuote,
                                  String customNull)
   {
      return quoterWorker(var, customQuote, customNull);
   }
   
   /**
    * Quoter function implementation for 3 parameters.
    * <p>
    * This puts the <code>String</code> value of the variable between
    * double quotes if the value is not <code>null</code> or replace it 
    * with '?' if the value is <code>null</code>.
    * <p>
    * Additionally, a parameter representing a custom quote character
    * is provided.  If not <code>null</code>, then the first character
    * of the given text will be used instead of the &quot; 
    * character.
    * <p>
    * A custom unknown value can also be provided.  If
    * not <code>null</code>, then the given text will be used instead
    * of the default ? <code>unknown</code> replacement to be used.
    *  
    * @param    var
    *           The <code>String</code> variable to operate upon.
    * @param    customQuote
    *           The custom <code>String</code> quoter variable.
    * @param    customNull
    *           The custom <code>String</code> <code>unknown</code> 
    *           replacement to be used.
    *
    * @return   The resulting quoted <code>character</code> value 
    *           of the function.
    */
   public static character quoter(String var,
                                  String customQuote,
                                  String customNull)
   {
      return quoterWorker(new character(var), customQuote, customNull);
   }
   
   /**
    * Tests a bit in the given position in the input array to determine if
    * that bit is a 0 or 1.  The input string is treated as a multibyte 
    * bitfield (a byte array). The first character of the string corresponds
    * to bit positions 0-7 (if it is non-DBCS and presumably in a DBCS
    * character the positions would be 0-15, although this is not known).  In
    * a non-DBCS charset, the second char would correspond to bits 8-15.
    * <p>
    * This feature supports an undocumented Progress built-in function called
    * _cbit().
    *
    * @param    data
    *           A string with 0 or more characters.
    * @param    index
    *           The bit position to test.
    *
    * @return   The return is always <code>false</code> if the empty string
    *           is input or if a negative index is specified.  For any string
    *           of > 0 bytes length, the return is <code>true</code> if the
    *           bit in the specified position is 1 and <code>false</code> if
    *           it is 0. 
    */
   public static logical testBitAt(character data, integer index)
   {
      if (data.isUnknown() || index.isUnknown())
      {
         return new logical();
      }
      
      return testBitAt(data.getValue(), index.intValue());
   }

   /**
    * Tests a bit in the given position in the input array to determine if
    * that bit is a 0 or 1.  The input string is treated as a multibyte 
    * bitfield (a byte array). The first character of the string corresponds
    * to bit positions 0-7 (if it is non-DBCS and presumably in a DBCS
    * character the positions would be 0-15, although this is not known).  In
    * a non-DBCS charset, the second char would correspond to bits 8-15.
    * <p>
    * This feature supports an undocumented Progress built-in function called
    * _cbit().
    *
    * @param    data
    *           A string with 0 or more characters.
    * @param    index
    *           The bit position to test.
    *
    * @return   The return is always <code>false</code> if the empty string
    *           is input or if a negative index is specified.  For any string
    *           of > 0 bytes length, the return is <code>true</code> if the
    *           bit in the specified position is 1 and <code>false</code> if
    *           it is 0. 
    */
   public static logical testBitAt(character data, int index)
   {
      if (data.isUnknown())
      {
         return new logical();
      }
      
      return testBitAt(data.getValue(), index);
   }

   /**
    * Tests a bit in the given position in the input array to determine if
    * that bit is a 0 or 1.  The input string is treated as a multibyte 
    * bitfield (a byte array). The first character of the string corresponds
    * to bit positions 0-7 (if it is non-DBCS and presumably in a DBCS
    * character the positions would be 0-15, although this is not known).  In
    * a non-DBCS charset, the second char would correspond to bits 8-15.
    * <p>
    * This feature supports an undocumented Progress built-in function called
    * _cbit().
    *
    * @param    data
    *           A string with 0 or more characters.
    * @param    index
    *           The bit position to test.
    *
    * @return   The return is always <code>false</code> if the empty string
    *           is input or if a negative index is specified.  For any string
    *           of > 0 bytes length, the return is <code>true</code> if the
    *           bit in the specified position is 1 and <code>false</code> if
    *           it is 0. 
    */
   public static logical testBitAt(String data, integer index)
   {
      if (index.isUnknown())
      {
         return new logical();
      }
      
      return testBitAt(data, index.intValue());
   }

   /**
    * Tests a bit in the given position in the input array to determine if
    * that bit is a 0 or 1.  The input string is treated as a multibyte 
    * bitfield (a byte array). The first character of the string corresponds
    * to bit positions 0-7 (if it is non-DBCS and presumably in a DBCS
    * character the positions would be 0-15, although this is not known).  In
    * a non-DBCS charset, the second char would correspond to bits 8-15.
    * <p>
    * This feature supports an undocumented Progress built-in function called
    * _cbit().
    *
    * @param    data
    *           A string with 0 or more characters.
    * @param    index
    *           The bit position to test.
    *
    * @return   The return is always <code>false</code> if the empty string
    *           is input or if a negative index is specified.  For any string
    *           of > 0 bytes length, the return is <code>true</code> if the
    *           bit in the specified position is 1 and <code>false</code> if
    *           it is 0. 
    */
   public static logical testBitAt(String data, int index)
   {
      // invalid/empty data or invalid index
      if (data == null || data.length() == 0 || index < 0)
      {
         return new logical(false);
      }
      
      byte[] bits    = data.getBytes();
      int    byteOff = index / 8;
      
      // index is too large 
      if (byteOff >= bits.length)
      {
         return new logical(false);
      }
      
      byte bitOff = (byte) (0xFF & (0x01 << (index % 8)));
      byte result = (byte) (0xFF & (bits[byteOff] & bitOff));
      
      return new logical(result != 0x00); 
   }

   /**
    * Convert a <code>double</code> into a character object using
    * the 'message' format for the <code>decimal</code> data type.
    *
    * @param    num
    *           The data to convert.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(double num)
   {
      return valueOf(new decimal(num));
   }
   
   /**
    * Convert an <code>int</code> into a character object using
    * the 'message' format for the <code>integer</code> data type.
    *
    * @param    num
    *           The data to convert.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(int num)
   {
      return valueOf(new integer(num));
   }
   
   /**
    * Convert a <code>String</code> into a character object using
    * the 'message' format for the <code>character</code> data type.
    *
    * @param    s
    *           The data to convert.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(String s)
   {
      return valueOf(new character(s));
   }
   
   /**
    * Convert a <code>boolean</code> into a character object using
    * the 'message' format for the <code>logical</code> data type.
    *
    * @param    b
    *           The data to convert.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(boolean b)
   {
      return valueOf(new logical(b));
   }

   /**
    * Convert a <code>rowid</code> into a character object using
    * the 'message' format for the <code>rowid</code> data type.
    * <p>
    * The returned value is guaranteed to have:
    * <ul>
    *    <li>'0x' prefix
    *    <li>only lowercase hexadecinal digits
    *    <li>even length, but not more that 16 digits
    * </ul>
    * <p>
    * As the actual internal storage is a java signed <code>long</code>
    * and the utility method <code>Long.toString()</code> will convert
    * negative values into string representations prefixed with '-' some
    * special code is needed to warp back the negative values and treat the
    * internal data as being an unsigned int.
    *
    * @param    r
    *           The data to convert.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(rowid r)
   {
      if (r.isUnknown())
      {
         // unknown value is represented by "?"
         return new character("?");
      }

      long val = r.getValue();
      boolean isNegative = val < 0;

      if (isNegative)
      {
         val -= Long.MIN_VALUE;
      }

      // lower-case hexadecimal representation of the 64 bit integer
      // Long.toString() uses lower-case characters for hex-digits
      String hex = Long.toString(val, 16);

      if (isNegative)
      {
         // as we wrap around the hex should be padded with 0
         // so negative values are 16 digits long:
         if (hex.length() != 16)
         {
            hex = "000000000000000".substring(0, 16 - hex.length()) + hex;
         }

         // fix the msf (the sign bit):
         int firstDigit = Character.digit(hex.charAt(0), 16);
         firstDigit += 0x08;
         // replace the first digit with the newly computed value:
         hex = Character.forDigit(firstDigit, 16) + hex.substring(1);
      }

      // make sure we have even length of digits after 0x prefix:
      return (hex.length() % 2 == 1) ?
         new character("0x0" + hex) :
         new character("0x" + hex);
   }
   
   /**
    * Convert any Progress compatible wrapper into a character object using
    * the 'message' format for the data type being converted.
    *
    * @param    bdt
    *           The data to convert.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(BaseDataType bdt)
   {
      if (bdt.isUnknown())
         return new character();

      return new character(bdt.toStringMessage());
   }   
   
   /**
    * Convert a <code>double</code> into a character object using
    * the given format for the <code>decimal</code> data type.
    *
    * @param    num
    *           The data to convert.
    * @param    fmt
    *           The format string to use.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(double num, String fmt)
   {
      return valueOf(new decimal(num), fmt);
   }
   
   /**
    * Convert an <code>int</code> into a character object using
    * the given format for the <code>integer</code> data type.
    *
    * @param    num
    *           The data to convert.
    * @param    fmt
    *           The format string to use.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(int num, String fmt)
   {
      return valueOf(new integer(num), fmt);
   }
   
   /**
    * Convert a <code>String</code> into a character object using
    * the given format for the <code>character</code> data type.
    *
    * @param    s
    *           The data to convert.
    * @param    fmt
    *           The format string to use.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(String s, String fmt)
   {
      return valueOf(new character(s), fmt);
   }
   
   /**
    * Convert a <code>boolean</code> into a character object using
    * the given format for the <code>logical</code> data type.
    *
    * @param    b
    *           The data to convert.
    * @param    fmt
    *           The format string to use.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(boolean b, String fmt)
   {
      return valueOf(new logical(b), fmt);
   }
   
   /**
    * Convert any Progress compatible wrapper into a character object using
    * the given format for the data type being converted.
    *
    * @param    bdt
    *           The data to convert.
    * @param    fmt
    *           The format string to use.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(BaseDataType bdt, String fmt)
   {
      if (bdt.isUnknown())
         return new character();

      return new character(bdt.toString(fmt));
   }
   
   /**
    * Convert a <code>double</code> into a character object using
    * the given format for the <code>decimal</code> data type.
    *
    * @param    num
    *           The data to convert.
    * @param    fmt
    *           The format string to use.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(double num, character fmt)
   {
      return valueOf(new decimal(num), fmt);
   }
   
   /**
    * Convert an <code>int</code> into a character object using
    * the given format for the <code>integer</code> data type.
    *
    * @param    num
    *           The data to convert.
    * @param    fmt
    *           The format string to use.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(int num, character fmt)
   {
      return valueOf(new integer(num), fmt);
   }
   
   /**
    * Convert a <code>String</code> into a character object using
    * the given format for the <code>character</code> data type.
    *
    * @param    s
    *           The data to convert.
    * @param    fmt
    *           The format string to use.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(String s, character fmt)
   {
      return valueOf(new character(s), fmt);
   }
   
   /**
    * Convert a <code>boolean</code> into a character object using
    * the given format for the <code>logical</code> data type.
    *
    * @param    b
    *           The data to convert.
    * @param    fmt
    *           The format string to use.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(boolean b, character fmt)
   {
      return valueOf(new logical(b), fmt);
   }
   
   /**
    * Convert any Progress compatible wrapper into a character object using
    * the given format for the data type being converted.
    *
    * @param    bdt
    *           The data to convert.
    * @param    fmt
    *           The format string to use.
    *
    * @return   The resulting string representation.
    */
   public static character valueOf(BaseDataType bdt, character fmt)
   {
      return valueOf(bdt, fmt.getValue());
   }
   
   /**
    * Converts a string (from a Progress source file) that represents a
    * Progress 4GL matching "pattern" to a traditional regular expression
    * which is suitable for a Java source file.  This allows a trivial
    * implementation of the matches operator using the
    * <code>String.matches</code> method.
    * <p>
    * In Progress 4GL, there are 2 escape characters '~' (tilde) and '\'
    * backslash.  In both Java and in regular expressions, the only escape
    * character is '\'.  In addition, since these strings come directly from
    * source files, the Progress runtime has not handled the escape sequences
    * on input as it normally would.  So although the sequences '~*' or '\*'
    * are normally the literal asterisk, the string must be coded to appear
    * as '~~*' or '\\*' in the source because the escape chars are normally 
    * dropped on input.  This method is written assuming that all of the
    * normal Progress conversions have already occurred by utilizing 
    * {@link #progressToJavaString} and {@link StringHelper#processEscapes}.
    * <p>
    * Once the standard Progress conversions have occurred, then this method
    * can be used and the following conversions will be additionally handled:
    * <pre>
    *    Progress                Regular Expression
    *    --------                ------------------
    *    ~.                      \\.
    *    \.                      \\.
    *    ~*                      \\*
    *    \*                      \\*
    *    ~~                      ~
    *    \~                      ~
    *    ~\                      \\\\
    *    \\                      \\\\
    *    *                       .*
    *    ~                       (removed or converted to \\ if the following char requires)
    *    \                       (removed or converted to \\ if the following char requires)
    * </pre>
    * <p>
    * The following characters in Progress have special meaning in the
    * extended regular expression syntax used in Java.  Each of these must
    * be properly escaped to ensure that their meaning is not accidentally
    * invoked by an input pattern that expects it to be matched literally:
    * <p>
    * <pre>
    *    Progress                Extended Regular Expression
    *    --------                ------------------
    *    ?                       \\\\?
    *    +                       \\\\+
    *    ^                       \\\\^
    *    $                       \\\\$
    *    |                       \\\\|
    *    (                       \\\\(
    *    )                       \\\\)
    *    [                       \\\\[
    *    ]                       \\\\]
    *    {                       \\\\{
    *    }                       \\\\}
    * </pre>
    * The following special Progress pattern constructs actually match those
    * of regular expressions and for this reason, these constructs are left
    * alone (not disturbed or modified):
    * <pre>
    *    Progress                Regular Expression
    *    --------                ------------------
    *    .                       .      
    * </pre>    
    * This conversion seems trivial, except that the backslash '\' character
    * must be handled properly since it needs to be doubled up for Java to
    * recognise one in a Java string and if two are needed in the regular
    * expression then one needs to double them up, twice.  This means that
    * in order to match a single \ in a regular expression, the Java string
    * will need \\\\ (4 backslashes).  In addition, since the
    * asterisk '*' character is a wildcard (it matches any number of any
    * character) in Progress, it must be properly replaced in the regular
    * expression (where .* is the unlimited wildcard).  Both Progress and
    * regular expressions use '.' to mean match any single character.
    *
    * @param   pattern
    *          Progress 4GL matching pattern.    
    *
    * @return  Converted regular expression.
    */
   public static String convertToRegEx(String pattern)
   {
      StringBuffer sb = new StringBuffer("");      
            
      int end = pattern.length();
      
      // convert the contents into the equivalent Java form
      for (int i = 0; i < end; i++) 
      {
         char current = pattern.charAt(i);
         char next;
         
         switch (current)
         {
            case '*':
               sb.append(".*");
               break;
            case '~':
            case '\\':
               // there must always be a following character
               i++;
               next = pattern.charAt(i);               
               
               // handle the special cases that don't have a one-to-one
               // correspondence in Java
               if (next == 'E')
               {
                  // the special case for the ESC char
                  sb.append("\\033");
               }
               else
               {
                  // only certain chars need to be AND in fact only these MAY
                  // BE escaped in Java so drop the preceding escape char
                  // unless the following char is in this list
                  if (next == 'b'  || next == 't'  || next == 'n' ||
                      next == 'f'  || next == 'r'  || next == '\''||
                      next == '\"' || next == '\\' || next == '.' ||
                      next == '*'  ||(next >= '0' && next <= '7'))
                  {
                     // we must convert the leading tilde into the Java
                     // escape char because all of these must be escaped in
                     // Java
                     sb.append('\\');
                  }
   
                  // insert our following char (warning: this is a bit funky
                  // for the octal escapes because we don't write the whole
                  // thing at once, but the following chars will all trip
                  // the default processing and be passed through so this
                  // works even if it isn't as obvious as we would like)
                  sb.append(next);                  
               }               
               break;            
            default:
               // certain chars need to have an escape inserted because they 
               // have no special meaning in Progress but do have a special
               // meaning in Java's extended regular expression syntax
               if (current == '?' || current == '+' || current == '^' ||
                   current == '$' || current == '|' || current == '(' ||
                   current == ')' || current == '{' || current == '}' ||
                   current == '[' || current == ']')
               {
                  // insert a leading escape char
                  sb.append("\\");
               }
   
               // not an escaped char or special case
               sb.append(current);
         }
      }
      
      return sb.toString();
   }

   /**
    * Converts a string (from within a character variable) that represents a
    * Progress 4GL matching "pattern" to a traditional SQL LIKE expression.
    * This variant is intended for runtime use.
    *
    * @param   pattern
    *          Progress 4GL matching pattern.    
    *
    * @return  Converted SQL LIKE expression, or unknown value if
    *          <code>pattern</code> represents the unknown value.
    *
    * @see     #convertToSQLLike(java.lang.String, boolean)
    */
   public static character convertToSQLLike(character pattern)
   {
      if (pattern.isUnknown())
      {
         return (new character());
      }
      
      return (new character(convertToSQLLike(pattern.getValue(), true)));
   }

   /**
    * Converts a string (from a Progress source file) that represents a
    * Progress 4GL matching "pattern" to a traditional SQL LIKE expression
    * which is suitable for a Java source file.  This allows a trivial
    * implementation of the matches operator using the SQL LIKE operator.
    * <p>
    * In Progress 4GL, there are 2 escape characters '~' (tilde) and '\'
    * backslash.  In both Java and in SQL, the only escape
    * character is '\'.  In addition, since these strings come directly from
    * source files, the Progress runtime has not handled the escape sequences
    * on input as it normally would.  So although the sequences '~*' or '\*'
    * are normally the literal asterisk, the string must be coded to appear
    * as '~~*' or '\\*' in the source because the escape chars are normally 
    * dropped on input.  This method is written assuming that all of the
    * normal Progress conversions have already occurred by utilizing 
    * {@link #progressToJavaString} and {@link StringHelper#processEscapes}.
    * <p>
    * Once the standard Progress conversions have occurred, then this method
    * can be used and the following conversions will be additionally handled:
    * <pre>
    *    Progress       SQL LIKE
    *    --------       -----------------
    *    ~.             .
    *    \.             .
    *    ~*             *
    *    \*             *
    *    ~~             ~
    *    \~             ~
    *    ~\             \\
    *    .              _
    *    *              %
    *    %              \\% (runtime); \\\\% (hardcoded)
    *    _              \\_ (runtime); \\\\_ (hardcoded)
    *    ~              (removed or converted to \\ if the following char requires)
    *    \              (removed or converted to \\ if the following char requires)
    * </pre>
    * This conversion seems trivial, except that the backslash '\' character
    * must be handled properly since it needs to be doubled up for Java to
    * recognise one in a Java string.
    *
    * @param   pattern
    *          Progress 4GL matching pattern.
    * @param   runtimeString
    *          <code>true</code> if the target string is a runtime string,
    *          <code>false</code> if it is a hardcoded string into a source
    *          file.
    *
    * @return  Converted SQL LIKE expression.
    */
   public static String convertToSQLLike(String pattern, boolean runtimeString)
   {
      StringBuffer sb = new StringBuffer("");      
            
      int end = pattern.length();
      
      // convert the contents into the equivalent Java form
      for (int i = 0; i < end; i++) 
      {
         char current = pattern.charAt(i);
         char next;
         
         switch (current)
         {
            case '*':
               sb.append("%");
               break;
            case '.':
               sb.append("_");
               break;
            case '~':
            case '\\':
               // there must always be a following character
               i++;
               next = pattern.charAt(i);               
               
               // handle the special cases that don't have a one-to-one
               // correspondence in Java
               if (next == 'E')
               {
                  // the special case for the ESC char
                  sb.append("\\033");
               }
               else
               {
                  // only certain chars need to be AND in fact only these MAY
                  // BE escaped in Java so drop the preceding escape char
                  // unless the following char is in this list
                  if (next == 'b'  || next == 't'  || next == 'n' ||
                      next == 'f'  || next == 'r'  || next == '\''||
                      next == '\"' || (next >= '0' && next <= '7') ||
                      next == '\\')
                  {
                     // we must convert the leading tilde into the Java
                     // escape char because all of these must be escaped in
                     // Java
                     sb.append('\\');
                  }
   
                  // insert our following char (warning: this is a bit funky
                  // for the octal escapes because we don't write the whole
                  // thing at once, but the following chars will all trip
                  // the default processing and be passed through so this
                  // works even if it isn't as obvious as we would like)
                  sb.append(next);                  
               }               
               break;
            case '%':
               sb.append(runtimeString ? "\\%" : "\\\\%");
               break;
            case '_':
               sb.append(runtimeString ? "\\_" : "\\\\_");
               break;
            default:
               // not an escaped char or special case
               sb.append(current);
         }
      }
   
      return sb.toString();
   }

   /**
    * Converts a second operand of the "op1 BEGINS op2" function into
    * corresponding SQL LIKE expression. It is assumed that output string is a
    * runtime string.
    * The following conversions will be performed:
    *  %    to    \\% (runtime); \\\\% (hardcoded)
    *  _    to    \\_ (runtime); \\\\_ (hardcoded)
    *  \\   to    \\\\
    * Also % will be appended to the end of the string.
    *
    * @param  str
    *         Represents a second operand of the "op1 BEGINS op2" function.
    *
    * @return target SQL LIKE expression.
    */
   public static character convertToSQLBegins(character str)
   {
      if (str.isUnknown())
      {
         return (new character());
      }
   
      return (new character(convertToSQLBegins(str.getValue(), true)));
   }

   /**
    * Converts a second operand of the "op1 BEGINS op2" function into
    * corresponding SQL LIKE expression.
    * The following conversions will be performed:
    *  %    to    \\% (runtime); \\\\% (hardcoded)
    *  _    to    \\_ (runtime); \\\\_ (hardcoded)
    *  \\   to    \\\\
    * Also % will be appended to the end of the string.
    *
    * @param  str
    *         Represents a second operand of the "op1 BEGINS op2" function.
    * @param  runtimeString
    *         <code>true</code> if the target string is a runtime string,
    *         <code>false</code> if it is a hardcoded string into a source
    *         file.
    *
    * @return The corresponding SQL LIKE expression. 
    */
   public static String convertToSQLBegins(String str, boolean runtimeString)
   {
      StringBuffer sb = new StringBuffer("");
   
      int end = str.length();
   
      for (int i = 0; i < end; i++)
      {
         char current = str.charAt(i);
   
         switch (current)
         {
            case '%':
               sb.append(runtimeString ? "\\%" : "\\\\%");
               break;
            case '_':
               sb.append(runtimeString ? "\\_" : "\\\\_");
               break;
            case '\\':
               sb.append("\\\\"); // extra escaping is added for hardcoded 
                                  // strings into postprocessStringLiteral
               break;
            default:
               sb.append(current);
         }
      }
      sb.append("%");
      return sb.toString();
   }

   /**
    * Converts a second operand of the "op1 CONTAINS op2" function into
    * corresponding SQL LIKE expression.
    * 
    * Progress syntax for op2 is some kind of recursive expression like:
    * <expr>     := word | word* | (<or-expr>)
    * <or-expr>  := <and-expr> ( <or-op> <and-expr> )*
    * <and-expr> := <expr> (<and-op> <expr>)*
    * <or-op>    is one of: "^" "|" "!"
    * <and-op>   = "&"
    * 
    *
    * @param  str
    *         Represents a second operand of the "op1 CONTAINS op2" clause.
    *
    * @return The corresponding SQL LIKE expression. 
    */
   public static character convertToSQLContains(character str)
   {
      if (str.isUnknown())
      {
         return new character();
      }

      return new character(convertToSQLContains(str.getValue(), true));
   }

   /**
    * Converts a second operand of the "op1 CONTAINS op2" function into
    * corresponding SQL LIKE expression.
    * <p>
    * Progress syntax for op2 is some kind of recursive expression like:
    * <pre>
    * &lt;expr&gt;     := word | word"*" | "(" &lt;or-expr&gt; ")"
    * &lt;or-expr&gt;  := &lt;and-expr&gt; ( &lt;or-op&gt; &lt;and-expr&gt; )*
    * &lt;and-expr&gt; := &lt;expr&gt; (&lt;and-op&gt; &lt;expr&gt;)*
    * &lt;or-op&gt;    is one of: "^" "|" "!"
    * &lt;and-op&gt;   = "&"
    * </pre>
    * Spaces are not allowed.<p>
    * However, the DIRTY method only does the following: <ul>
    * <li>looks for words
    * <li>removes * from the end, if any
    * <li>replaces special characters (see {@link #convertToSQLBegins()})
    * <li>replaces the word with %word% in the original text
    * </ul>
    * This way, if no fancy expression is used the correct query is generated. Any more complex 
    * expression cannot be converted without recursive processing of the expression using the 
    * above definition.
    * <p>
    * A better algorithm would convert a clause like:
    * <pre>SomeField CONTAINS "word1&(word2*|word3)</pre>
    * into the more complex SQL form:
    * <pre>
    * ((SomeField LIKE "%word1%) AND ((SomeField LIKE "%word2%) OR(SomeField LIKE "%word3%)))
    * </pre>
    * The correct implementation has to use regular expressions together with SQL clause 
    * <code>SIMILAR TO</code>. However, I am not aware of how and if hibernate supports it. 
    *
    * @param  str
    *         Represents a second operand of the "op1 CONTAINS op2" clause.
    * @param  runtimeString
    *         <code>true</code> if the target string is a runtime string,
    *         <code>false</code> if it is a hardcoded string into a source
    *         file.
    *
    * @return A substitution SQL LIKE expression. 
    */
   public static String convertToSQLContains(String str, boolean runtimeString)
   {
      StringBuffer sb = new StringBuffer();
      Pattern pattern = Pattern.compile("[\\s\\(\\)\\^\\|\\!]");
      Matcher m = pattern.matcher(str);
      String word;
      int wordStart = 0;
      while (m.find()) 
      {
         String tok = m.group();
         int beg = m.start();
         int end = m.end();

         word = str.substring(wordStart, beg);
         if (!word.isEmpty()) 
         {
            if (word.endsWith("*"))
            {
               word = word.substring(0,  word.length()-1);
            }
            sb.append('%');
            sb.append(character.convertToSQLBegins(word, runtimeString));
         }
         sb.append(tok);
         wordStart = end;
      }

      word = str.substring(wordStart);
      if (!word.isEmpty())
      {
         if (word.endsWith("*"))
         {
            word = word.substring(0,  word.length()-1);
         }
         sb.append('%');
         sb.append(character.convertToSQLBegins(word, runtimeString));
      }

      return sb.toString();
   }

   /**
    * Removes a set of enclosing double quotes (first and last characters)
    * and converts any doubled up instances (2 instances sequentally) of
    * double quotes into a single instance of double quotes. This reverses
    * the reversable portion of the export format into a normal string as
    * Java would expect to see it.
    *
    * @param    export
    *           A string with the contents in export format as created by
    *           {@link #toStringExport}.
    *
    * @return   A valid Java string.
    */
   public static String fromExportString(String export)
   {
      // this will always work as long as the input string is valid
      String contents = export.substring(1, export.length() - 1);
      
      // remove any doubling up of double quote chars and convert to a single
      // instance
      return contents.replaceAll("\"\"", "\"");
   }

   /**
    * Converts a Progress 4GL compatible string to the equivalent Java source 
    * string, by stripping off enclosing quotes/string options and converting
    * the contents (including rewriting all escape sequences). Note that the
    * output format is designed to be written to a Java source file, which
    * requires a different approach than writing a string that can be used 
    * at runtime.
    * <p>
    * Escaped characters (using either the tilde or the backslash character)
    * and special formations will be converted to native representations and
    * any characters that need to be escaped will be so escaped.  In
    * particular, almost all usage of the backslash is compatible between
    * Progress and Java, so only those constructs that must be different
    * will be handled differently.
    * <p>
    * The following conversions will be made:
    * <p>
    * <pre>
    * ~b       to   '\b'   - backspace
    * ~t       to   '\t'   - horizontal tab
    * ~n       to   '\n'   - linefeed
    * ~f       to   '\f'   - formfeed
    * ~r       to   '\r'   - carriage return
    * ~&quot;       to   '\&quot;'   - double quote
    * ~&apos;       to   '\&apos;'   - single quote
    * ~\       to   '\\'   - backslash
    * ~~       to   '~'    - single tilde
    * ~{       to   '{'    - left curly brace
    * ~E       to   '\033' - the ASCII escape char
    * ~nnn     to   '\nnn' - octal escape sequence (nnn is a 3 digit octal number)
    * &apos;&apos;       to   '\&apos;'   - single quote (only when inside a
    *                                       string that was enclosed by single
    *                                       quotes)
    * &quot;&quot;       to   '\&quot;'   - double quote (only when inside a
    *                                       string that was enclosed by double
    *                                       quotes)
    * \~       to   '~'    - single tilde 
    * \\       to   '\\'   - backslash
    * \{       to   '{'    - left curly brace
    * \E       to   '\033' - the ASCII escape char
    * </pre>
    * <p>
    * Any other escape sequence that is unrecognized will result in the
    * tilde or backslash being dropped and the following character output
    * unchanged which is the same behavior as Progress.
    * <p>
    * The final result when output will be output inside enclosing double
    * quotes but there is no need for those quotes inside the AST form.    
    *
    * @param    progress
    *           The Progress compatible string as parsed from the source
    *           file.
    *
    * @return   A valid Java string that can be written as a string literal.
    */
   public static String progressToJavaString(String progress)
   throws IllegalArgumentException
   {
      return progressToJavaString(progress, false);
   }

   /**
    * Converts a Progress 4GL compatible string to the equivalent Java source 
    * string, by stripping off enclosing quotes/string options and converting
    * the contents (including rewriting all escape sequences). Note that the
    * output format is designed to be written to a Java source file, which
    * requires a different approach than writing a string that can be used 
    * at runtime.
    * <p>
    * Escaped characters (using either the tilde or the backslash character)
    * and special formations will be converted to native representations and
    * any characters that need to be escaped will be so escaped.  In
    * particular, almost all usage of the backslash is compatible between
    * Progress and Java, so only those constructs that must be different
    * will be handled differently.
    * <p>
    * The following conversions will be made:
    * <p>
    * <pre>
    * ~b       to   '\b'   - backspace
    * ~t       to   '\t'   - horizontal tab
    * ~n       to   '\n'   - linefeed
    * ~f       to   '\f'   - formfeed
    * ~r       to   '\r'   - carriage return
    * ~&quot;       to   '\&quot;'   - double quote
    * ~&apos;       to   '\&apos;'   - single quote
    * ~\       to   '\\'   - backslash
    * ~~       to   '~'    - single tilde
    * ~{       to   '{'    - left curly brace
    * ~E       to   '\033' - the ASCII escape char
    * ~nnn     to   '\nnn' - octal escape sequence (nnn is a 3 digit octal number)
    * &apos;&apos;       to   '\&apos;'   - single quote (only when inside a
    *                                       string that was enclosed by single
    *                                       quotes)
    * &quot;&quot;       to   '\&quot;'   - double quote (only when inside a
    *                                       string that was enclosed by double
    *                                       quotes)
    * \~       to   '~'    - single tilde 
    * \\       to   '\\'   - backslash
    * \{       to   '{'    - left curly brace
    * \E       to   '\033' - the ASCII escape char
    * </pre>
    * <p>
    * Any other escape sequence that is unrecognized will result in the
    * tilde or backslash being dropped and the following character output
    * unchanged which is the same behavior as Progress.
    * <p>
    * The final result when output will be output inside enclosing double
    * quotes but there is no need for those quotes inside the AST form.    
    *
    * @param    progress
    *           The Progress compatible string as parsed from the source
    *           file.
    *
    * @return   A valid Java string that can be written as a string literal.
    */
   public static String progressToJavaString(String progress, boolean windows)
   throws IllegalArgumentException
   {  
      if (progress == null)
         throw new IllegalArgumentException("Invalid input: <null>.");
      
      // are we starting with ' or "
      char lead = progress.charAt(0);
      
      String contents = null;
      
      try
      {   
         // this will always work as long as the input string is valid
         contents = progress.substring(1, progress.lastIndexOf(lead));
      }
      
      catch (IndexOutOfBoundsException iob)
      {
         throw new IllegalArgumentException("Missing closing quote (" +
                                            lead + ").");
      }
      
      StringBuffer sb = new StringBuffer();
      
      int end = contents.length();
      
      // quick exit for the empty string
      if (end == 0)
      {
         return "";
      }
            
      // by now the enclosing quotes are gone, as are any string options
      // so we must convert the contents into the equivalent Java form
      for (int i = 0; i < end; i++) 
      {
         char current = contents.charAt(i);
         char next;
         
         switch (current)
         {
            case '\'':
            case '\"':
               // does the quote type match our enclosing quotes or not?
               if (current == lead)
               {
                  // this should only occur if we have "doubled up" quotes
                  // and we need to "consume" the extra character
                  i++;
                  next = readChar(contents, i);
                  
                  // check to ensure we aren't smoking something here
                  if (next != current)
                     throw new IllegalArgumentException("Malformed string.");
               }
               
               // escape the double quote as in Progress it safely stands 
               // alone
               if (current == '\"')
               {
                  sb.append('\\');
               }
               
               sb.append(current);
               break;
            case '~':
            case '\\':
               // For Windows-target we have to make special processing for
               // '\' characters in path related strings
               if (windows && current == '\\')
               {
                  sb.append("\\\\");
                  break;
               }
   
               // there must always be a following character
               i++;
               next = readChar(contents, i);
               
               // handle the special cases that don't have a one-to-one
               // correspondence in Java
               if (next == 'E')
               {
                  // the special case for the ESC char
                  sb.append("\\033");
               }
               else
               {
                  // only certain chars need to be AND in fact other than the
                  // single quote (which we deliberately omit here) only 
                  // these MAY BE escaped in Java so drop the preceding
                  // escape char unless the following char is in this list
                  if (next == 'b'  || next == 't'  || next == 'n'  ||
                      next == 'f'  || next == 'r'  || next == '\"' || 
                      next == '\\' || (next >= '0' && next <= '7'))
                  {
                     // we must convert the leading tilde into the Java
                     // escape char because all of these must be escaped in
                     // Java
                     sb.append('\\');
                  }
   
                  // insert our following char (warning: this is a bit funky
                  // for the octal escapes because we don't write the whole
                  // thing at once, but the following chars will all trip
                  // the default processing and be passed through so this
                  // works even if it isn't as obvious as we would like)
                  sb.append(next);                  
               }               
               break;            
            default:
               // not an escaped char or special case
               sb.append(current);
         }
      }
      
      return sb.toString();
   }

   /**
    * Return a string which contains the contents of the given string up to 
    * (but not including) the first <code>null</code> character (encoded as a
    * Progress escape sequence) and for that <code>null</code> character and 
    * all following characters the resulting string will contain spaces. 
    * <p>
    * The resulting length of the string (in terms of the number of
    * characters that the string will include at runtime) will NOT change,
    * although the length of the string representation (as a source code
    * representation stored in a Java string) may be reduced as the
    * <code>null</code> character's escape sequence (which would require 4
    * characters to represent as in <code>~000</code>) and any following
    * escaped sequences will be converted to a space which only consumes
    * a single character position.
    * <p>
    * The enclosing single or double quote characters are maintained (not
    * overwritten).
    * <p>
    * This processing is needed by the Progress preprocessor and other
    * conversion processing of string literals.    
    * <p>
    * This transformation is for Progress compatible string literals, NOT for
    * runtime strings.  If a runtime string conversion is needed, please use
    * {@link #javaSpacifyNull}.  Both forms of the <code>null</code>
    * character escape sequence are handled.
    * <p>
    * As an example, a Progress string such as:
    * <p>
    * <pre>
    * Input                               Output
    * ----------------------------------  ----------------------------------
    * "This has a ~000 (null) char."      "This has a               "      
    * 'This has a ~000 (null) char.'      'This has a               '
    * "~000"                              " "
    * '~000'                              ' '
    * "\000"                              " "
    * '\000'                              ' '
    * </pre>
    *
    * @param    data
    *           A Progress source code representation of a string literal
    *           which is the input string to be transformed.
    * @param    unixEsc
    *           <code>true</code> if the backslash character should be
    *           treated as an escape character.
    *
    * @return   The "space-ified" output string (still valid as a Progress
    *           string literal that could be output into source code).           
    */
   public static String progressSpacifyNull(String data, boolean unixEsc)
   {
      StringBuffer sb = new StringBuffer();
      
      // put the leading single or double quote char into the output buffer 
      char quote = data.charAt(0);
      sb.append(quote);
      
      // only scan up to the trailing quote char
      int end = data.length() - 1;
      
      boolean eat = false;
      
      // loop through the contents
      for (int i = 1; i < end; i++) 
      {
         char current = data.charAt(i);
         char next;
         
         if (current == '~' || (current == '\\' && unixEsc))
         {
            if (!eat)
            {
               // flow through mode
               
               // check for the null escape sequence
               String esc = ((i + 4) <= end) ? data.substring(i + 1, i + 4)
                                             : null;
               
               if (esc != null && esc.equals("000"))
               {
                  // the first null, shift into space-ify mode
                  eat = true;
                  i += 3;
                  sb.append(' ');
               }
               else
               {
                  // some other escape sequence, allow it to flow through
                  sb.append(current);
                  
                  // there must always be a following character, if it is
                  // a doubled escape char, we must handle it here to avoid 
                  // dropping back through this code since it is not the
                  // opening of a new escape sequence (since all following
                  // chars can be handled the same way we just go ahead
                  // no matter which one it is)
                  i++;
                  sb.append(readChar(data, i));
               }
            }
            else
            {
               // space-ify mode (eat the full escape sequence here)
               
               // output the replacement (always 1 char)
               sb.append(' ');
               
               // there must always be at least 1 following character
               i++;
               next = readChar(data, i);
               
               if (next == 'b'  || next == 't'  || next == 'n'  ||
                   next == 'f'  || next == 'r'  || next == '\"' || 
                   next == '\\' || next == '~'  || next == 'E')
               {
                  // these are only 1 char in size, so we are done
               }
               else if (next >= '0' && next <= '7')
               {
                  // eat the next 2 chars too because this is an octal
                  // escape sequence
                  i += 2;
               }               
            }
         }
         else
         {
            sb.append(eat ? ' ' : current);
         }
      }
      
      // add the trailing (close) quote char
      sb.append(data.charAt(end));
      
      return sb.toString();      
   }

   /**
    * Return a string which contains the contents of the given string up to 
    * (but not including) the first <code>null</code> character (ASCII 0x00)
    * and for that <code>null</code> character and all following characters
    * the resulting string will contain spaces. 
    * <p>
    * The resulting length of the output string will be the same as the
    * length of the input string.
    * <p>
    * This processing is needed by some operators and/or functionality that
    * process <code>character</code> data.
    * <p>
    * This transformation is for runtime strings, NOT for Progress compatible
    * string literals.  If a source code conversion is needed, please use
    * {@link #progressSpacifyNull}.
    * <p>
    * As an example, a string such as:
    * <p>
    * <pre>
    * Input                               Output
    * ----------------------------------  ----------------------------------
    * "This has a \0 (null) char."        "This has a               "      
    * "\0"                                " "
    * </pre>
    *
    * @param    data
    *           The input string (not valid for encoding in a source file).
    *
    * @return   The "space-ified" output string (still not valid for
    *           encoding in a source file).           
    */
   public static String javaSpacifyNull(String data)
   {
      StringBuffer sb = null;
      
      int idx = data.indexOf('\0');
      
      if (idx != -1)
      {
         sb = new StringBuffer(data.substring(0, idx));
         StringHelper.repeatChar(' ', (data.length() - idx), sb);
      }
      
      return sb == null ? data : sb.toString();
   }

   /**
    * Return a string which contains the contents of the given string up to 
    * (but not including) the first <code>null</code> character (ASCII 0x00),
    * the resulting string is truncated at the first <code>null</code>.
    * <p>
    * As an example, a string such as:
    * <p>
    * <pre>
    * Input                               Output
    * ----------------------------------  ----------------------------------
    * "This has a \0 (null) char."        "This has a "      
    * "\0"                                ""
    * </pre>
    *
    * @param    data
    *           The input string (not valid for encoding in a source file).
    *
    * @return   The truncated output string (still not valid for
    *           encoding in a source file).           
    */
   public static String javaTruncateNull(String data)
   {
      StringBuffer sb = null;
      
      int idx = data.indexOf('\0');
      
      if (idx != -1)
      {
         sb = new StringBuffer(data.substring(0, idx));
      }
      
      return sb == null ? data : sb.toString();
   }

   /**
    * Convert the Progress 4GL style character type format string into a 
    * form that is easier to process.  All non-escaped 'A(n)' forms are
    * expanded into n instances of A.  All recognized (non-fill) format
    * characters are lowercased in the output to make comparisons easier.
    *
    * @param    fmt
    *           The Progress 4GL format string for a character data type.
    *
    * @return   The simplified/expanded format string.
    */
   public static String preprocessFormatString(String fmt)
   {
      StringBuffer sb = new StringBuffer();
      
      int  len = fmt.length();
      char c   = (char) 0;
      
      char    last;
      boolean nonFill = false;
      
      for (int i = 0; i < len; i++) 
      {
         last = c;
         c    = fmt.charAt(i);
         
         switch (c)
         {
            case 'x':
            case 'X':
            case 'n':
            case 'N':
            case 'a':
            case 'A':
            case '!':
            case '9':
               if (last != '~')
               {
                  // these are recognized valid 
                  nonFill = true;
                  sb.append(Character.toLowerCase(c));
               }
               else
               {
                  // any fill chars just go into the output unchanged
                  nonFill = false;
                  sb.append(c);
               }
               break;
            case '(':
               if (last != '~' && nonFill)
               {  
                  // find the end of the parens
                  int j = fmt.indexOf(')', i + 1);  
                  
                  if (j < 0)
                  {
                     String spec = "Character format %s is incomplete";
                     String err  = String.format(spec, fmt);
                     
                     ErrorManager.recordOrThrowError(164, err);
                     
                     // this is the safest thing to return (any results
                     // should be undone anyway)
                     return "";                     
                  }
                  
                  // how many repeats do we have?
                  int num = Integer.parseInt(fmt.substring(i+1, j));
                  
                  // get our last char
                  char ch = Character.toLowerCase(last);
                  
                  // create and add the resulting string which takes advantage
                  // of the fact that the last char is already in the buffer
                  // and so we only need to repeat (num - 1) times
                  sb.append(StringHelper.repeatChar(ch, num - 1));
                  
                  // move our index up to the end of the section, it will
                  // be incremented past/tested against the string length
                  // by the for loop itself
                  i = j;
               }
               else
               {
                  if (last == '~')
                  {
                     // replace the tilde with this char
                     sb.setCharAt(sb.length() - 1, c);
                  }
                  else
                  {
                     // we must be following a fill char so this is a fill
                     // char too
                     sb.append(c);
                  }
               }
               nonFill = false;
               break;
            default:
               // any fill chars just go into the output unchanged
               nonFill = false;
               sb.append(c);
         }
      }
      
      return sb.toString();
   }

   /**
    * Performs postprocessing of the string literals into HQL expressions.
    * The following conversions will be performed:
    *  '    to    ''
    *  \\   to    \\\\ (runtime); \\\\\\\\ (hardcoded)  
    *
    * @param  text
    *         The string literal to process (without enclosing quotes).
    * @param  runtimeString
    *         <code>true</code> if the target string is a runtime string,
    *         <code>false</code> if it is a hardcoded string into a source
    *         file.
    *
    * @return processed string literal.
    */
   public static String postprocessStringLiteral(String text,
                                                 boolean runtimeString)
   {
      StringBuffer sb = new StringBuffer("");
      int end = text.length();

      for (int i = 0; i < end; i++)
      {
         char current = text.charAt(i);

         switch (current)
         {
            case '\'':
               sb.append("\'\'");
               break;

            case '\\':
               if (runtimeString)
               {
                  sb.append("\\\\");
               }
               else
               {
                  i++;
                  char next = text.charAt(i);
                  if (next == '\\')
                  {
                     sb.append("\\\\\\\\");
                  }
                  else
                  {
                     sb.append(current);
                     sb.append(next);
                  }
                  break;
               }

            default:
               sb.append(current);
         }
      }

      return sb.toString();
   }
   
   /**
    * Returns a clone of the current instance which has all <code>null</code>
    * characters removed from its data (in particular, the data is truncated 
    * starting at any <code>null</code> character).
    *
    * @return   An instance which has no embedded <code>null</code>
    *           characters.
    */
   public character safeDuplicate()
   {
      character ch = (character) duplicate();
      
      if (!unknown)
      {
         ch.value = safeValue();
      }
      
      return ch;
   }

   /**
    * Returns the instance's value safely truncated at any <code>null</code>
    * byte.
    *
    * @return   A value with no <code>null</code> bytes.
    */
   protected String safeValue()
   {
      return javaTruncateNull(value);
   }

   /**
    * Return the default display format string for this type.
    *
    * @return   The default format string.
    */
   public String defaultFormatString()   
   {
      return defaultFormat;
   }

   /**
    * Calculate the length of a formatted value (the text form of the value)
    * using this format string.
    *
    * @param    fmt
    *           The format string.
    *
    * @return   The length in characters of the any resulting formatted
    *           value.
    */
   public int formatLength(String fmt)
   {
      if (fmt == null)
      {
         fmt = defaultFormatString();
      }
      
      return super.formatLength(preprocessFormatString(fmt));
   }

   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.  
    */
   public BaseDataType duplicate()
   {
      return new character(this);
   }

   /**
    * Creates a new instance of the same type that represents the 
    * default initialized value.
    *
    * @return   An instance that represents the default value.
    */
   public BaseDataType instantiateDefault()   
   {
      return new character("");
   }

   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return   The 'message' formatted string.
    */
   public String toStringMessage()   
   {
      // quick-out when unknown
      if (unknown)
         return "?";
      
      return javaTruncateNull(getValue());
   }

   /**
    * Creates a string representation of the instance data using the 'export'
    * format.  For the character data type the entire string is returned
    * enclosed in a pair of double quotes.  If the instance represents the
    * <code>unknown value</code>, a '?' will be returned.  Any instances of
    * double quotes inside the string will be doubled up on output.
    *
    * @return   The 'export' formatted string.
    */
   public String toStringExport()   
   {
      // quick-out when unknown (no user override for that case)
      if (unknown)
         return "?";
      
      // open the string
      StringBuffer sb = new StringBuffer("\"");
      
      // truncate at any embedded null
      String txt = javaTruncateNull(value);
      
      // add the contents, doubling up all instance of double quote chars
      sb.append(txt.replaceAll("\"", "\"\""));
      
      // close the string
      sb.append("\"");
      
      return sb.toString();
   }

   /**
    * Creates a string representation of the instance data with the default
    * Progress 4GL format of 'x(8)'. If the instance represents the
    * <code>unknown value</code>, a '?' will be returned.
    *
    * @return   The formatted string.
    */
   public String toString()   
   {
      return toString(null);
   }

   /**
    * Creates a string representation of the instance data using the user-
    * specified Progress 4GL format string.  If the passed-in format is
    * <code>null</code>, then a default format string of 'x(8)' will be
    * used.  If the instance represents the <code>unknown value</code>,
    * a '?' will be returned.
    *
    * @param    fmt
    *           The Progress 4GL format string for a character data type.
    *
    * @return   The formatted string.
    */
   public String toString(String fmt)   
   {
      // quick-out when unknown (no user override for that case)
      if (unknown)
         return "?";
      
      // implement 'default' behavior 
      if (fmt == null)
      {
         fmt = defaultFormat;
      }
      
      // make the format string easier to process
      String format = preprocessFormatString(fmt);
      
      // handle embedded nulls and trim trailing whitespace off the value
      String clean = StringHelper.safeTrimTrailing(javaTruncateNull(value));
      
      // build the output string and return it
      return processDataString(clean, format);
   }

   /**
    * Convert the input data into an output string that is formatted as
    * specified by the Progress 4GL style character type format string.
    * <p>
    * Recognized format specifier characters are x, X, n, N, a, A, 9 and
    * !. The presence of any unescaped (with tilde) instance of one of these
    * characters in the format string means that should there be a
    * corresponding character (at the same format specifier index) in the
    * data string, it will be copied to the output string at that same
    * position.  No transformation of the data is done except for the !
    * specifier which causes the data character to be uppercased before
    * output.
    * <p>
    * Any non-recognized format specifier and any escaped character
    * (including escaped format specifiers) are considered user-defined
    * "fill" characters.  These characters are copied from the format
    * string (not the data string) into the output string.  They are copied
    * unchanged.  Any unescaped tildes are dropped (not copied to the
    * output).
    * <p>
    * The input data will be copied to a location in the output string based
    * on the simple index position in the input data AND the relative position
    * of each format specifier.  The presence of fill characters does not
    * change the relative position of format specifier characters. This means
    * that the 1st character in the input string will be copied on output
    * to the location of the 1st format specifier character in the format
    * string, which may not be the first position in the output string (if
    * there are fill characters in the format string before the first format
    * specifier). The 2nd character in the input string will be copied on
    * output to the location of the 2nd format specifier, which may not
    * immediately follow the 1st data character in the output string if there
    * are intervening fill characters between the first and second format
    * specifiers.
    * <p>
    * If there is less data characters (in the input string) than format
    * specifier characters (in the format string), then a ' ' (space) will
    * be output for each trailing format specifier position which has no
    * data to be copied.  This is a kind of padding.
    * <p>
    * If there are more data characters (in the input string) than format
    * specifier characters (in the format string), then the output string
    * will be truncated, not copying those extra data characters (in the
    * input string) to the output string.
    * <p>
    * The output string length will always be equivalent to the size of the
    * format string less any dropped tilde escape characters.
    *
    * @param    data
    *           The input string.
    * @param    fmt
    *           The Progress 4GL format string for a character data type.
    *
    * @return   The output (formatted) string.
    */
   private String processDataString(String data, String fmt)
   {
      StringBuffer sb = new StringBuffer();
      
      int  dlen = data.length();
      int  flen = fmt.length();
      char c    = (char) 0;
      
      char last;
                                                               
      // iterate through all characters in the format string (the result
      // will be the same length as the *real* number of characters in the
      // format string --> unescaped tildes don't count) 
      for (int j = 0, i = 0; i < flen; i++) 
      {
         // remember the last character to track if it is a tilde 
         last = c;
         
         // get next format string character
         c = fmt.charAt(i);
         
         switch (c)
         {
            case '~':
               if (last == '~')
               {
                  // we leave a single tilde in the output and reset our
                  // character to ensure that tilde processing is reset
                  sb.append(c);
                  c = (char) 0;
               }
               break;
            case 'x':
            case 'X':
            case 'n':
            case 'N':
            case 'a':
            case 'A':
            case '9':
            case '!':
               // these are recognized valid format characters
               if (last == '~')
               {
                  // this was escaped so it is a fill char, this just goes
                  // into the output unchanged
                  sb.append(c);
               }
               else
               {
                  if (j < dlen)
                  {
                     // there is still data to output
                     
                     // only the ! causes the data to be transformed before
                     // output, all others cause the unchanged char to be
                     // output
                     char out = data.charAt(j);
                     
                     if (c == '!')
                        out = Character.toUpperCase(out);
                     
                     sb.append(out);
                     j++;
                  }
                  else
                  {
                     // no more data, append a space to pad the end
                     sb.append(' ');
                  }
               }
               break;
            default:
               // must be an unescaped fill char, add to the output unchanged
               sb.append(c);
         }
      }
      
      return sb.toString();
   }

   /**
    * Test for certain failure conditions and if OK return the indexed
    * character, otherwise throw an exception with useful debugging
    * information.
    *
    * @param    input
    *           The string to read from.
    * @param    index
    *           The 0-based index into the string, from which to read the 
    *           character.
    *
    * @return   The character at the specified index.
    *
    * @throws   IllegalArgumentException
    *           If the index is out of bounds.
    */
   private static char readChar(String input, int index)
   throws IllegalArgumentException
   {
      char next;
      
      if (index == input.length())
      {
         throw new IllegalArgumentException("Unexpected end of string.");
      }
      
      try
      {
         next = input.charAt(index); 
      }
      
      catch (IndexOutOfBoundsException iob)
      {
         throw new IllegalArgumentException("Invalid index (" + index +
                                            " in string \"" + input + "\").");
      }
      
      return next; 
   }
   
   /**
    * This is the core quoter function implementation.  This puts the text 
    * valueof the variable between double quotes if the value is not unknown,
    *  or it will replace it with the <code>?</code> character if the value is
    * <code>unknown</code>. If the variable is a <code>character</code> type,
    * any contained double quote character will be doubled (another double
    * quote character will be added next to it - Progress behaviour). 
    * <p>
    * The caller can provide replacement characters to use instead of the 
    * double quote character (&quot;) and instead of the default 
    * <code>unknown</code> character <code>?</code>.
    * 
    * @param    var
    *           The <code>BaseDataType</code> variable to operate upon.
    * @param    customQuote
    *           The custom quoter character to be used.If <code>null</code> 
    *           default quoter value is used.
    * @param    customNull
    *           The custom unknown text replacement to be used. If 
    *           <code>null</code>  default value is used.
    *
    * @return   The resulting quoted string or <code>unknown value</code>, 
    *           if the input is <code>unknown</code>.
    */
   private static character quoterWorker(BaseDataType var,
                                       String customQuote,
                                       String customNull)
   {
      // custom null character
      String nullVal = (customNull != null)  ? customNull 
                                             : "?";

      // custom quoter character
      // Progress behaviour if given custom character that 
      // isn't " or ' it will take the first letter as the custom quote char
      String quoteVal = (customQuote != null) ? customQuote.substring(0, 1) 
                                              : "\"";

      // for character values that contain the quote character
      // it will be doubled where met inside the text
      String in = quoteVal;
      String out = quoteVal + quoteVal;
      String result = nullVal;

      if (!var.isUnknown())
      {
         if (var instanceof character)
         {
            character charVar = (character) var;
            result = TextOps.concat(quoteVal,
                            charVar.getValue().replaceAll(in, out),
                            quoteVal).getValue();
         }
         else if (var instanceof date)
         {
            date dateVar = (date) var;
            // date format result of quoter is 99/99/9999
            result = TextOps.concat(quoteVal,
                            dateVar.toString("99/99/9999"),
                            quoteVal).getValue();
         }
         else
         {
            // the rest of the data type format result of quoter
            // match the default message format (integer, decimal, logical etc)
            result = TextOps.concat(quoteVal, 
                            var.toStringMessage(), 
                            quoteVal).getValue();
         }
      }

      return new character(result);
   }
   
   /**
    * Encapsulates the logic of storing and executing a set of tests using
    * a minimum set of input (2 string operands and a set of boolean results
    * that are to be expected).
    */
   private static class TestCase
   {
      /** Operand 1. */
      character op1 = null;
      
      /** Operand 2. */
      character op2 = null;
      
      /** Expected test results for begins. */
      boolean begins;
      
      /** Expected test results for equals. */
      boolean eq;
      
      /** Expected test results for notEquals. */
      boolean ne;
      
      /** Expected test results for isGreaterThan. */
      boolean gt;
      
      /** Expected test results for isLessThan. */
      boolean lt;
      
      /** Expected test results for isGreaterThanOrEqual. */
      boolean ge;
      
      /** Expected test results for isLessThanOrEqual. */
      boolean le;
      
      /** Expected test results for matches. */
      boolean matches;
      
      /** Expected test results for case-sensitive begins. */
      boolean cBegins;
      
      /** Expected test results for case-sensitive equals. */
      boolean cEq;
      
      /** Expected test results for case-sensitive notEquals. */
      boolean cNe;
      
      /** Expected test results for case-sensitive isGreaterThan. */
      boolean cGt;
      
      /** Expected test results for case-sensitive isLessThan. */
      boolean cLt;
      
      /** Expected test results for case-sensitive isGreaterThanOrEqual. */
      boolean cGe;
      
      /** Expected test results for case-sensitive isLessThanOrEqual. */
      boolean cLe;
      
      /** Expected test results for case-sensitive matches. */
      boolean cMatches;
      
      TestCase(String op1 , String op2     , boolean begins , boolean eq,
               boolean ne , boolean gt     , boolean lt     , boolean ge,
               boolean le , boolean matches, boolean cBegins, boolean cEq,
               boolean cNe, boolean cGt    , boolean cLt    , boolean cGe,
               boolean cLe, boolean cMatches)
      {
         // simulate the resulting runtime strings
         String s1 = progressToJavaString("'" + op1 + "'");
         String s2 = progressToJavaString("'" + op2 + "'");
         
         this.op1      = new character(StringHelper.processEscapes(s1));
         this.op2      = new character(StringHelper.processEscapes(s2));
         this.begins   = begins;
         this.eq       = eq;
         this.ne       = ne;
         this.gt       = gt;
         this.lt       = lt;
         this.ge       = ge;
         this.le       = le;
         this.matches  = matches;
         this.cBegins  = cBegins;
         this.cEq      = cEq;
         this.cNe      = cNe;
         this.cGt      = cGt;
         this.cLt      = cLt;
         this.cGe      = cGe;
         this.cLe      = cLe;
         this.cMatches = cMatches;
      }
      
      boolean test()
      {
         boolean flag = true;
         
         if (TextOps._begins(op1, op2) != begins)
         {
            flag = false;
            System.out.println("begins failure for \"" + op1 + "\" and \"" +
                               op2 + "\"");
         }
         op1.setCaseSensitive(true);
         if (TextOps._begins(op1, op2) != cBegins)
         {
            flag = false;
            System.out.println("begins(c) failure for \"" + op1 +
                                "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(false);
         if (CompareOps.equals(op1, op2) != eq)
         {
            flag = false;
            System.out.println("isEqual failure for \"" + op1 + "\" and \"" +
                               op2 + "\"");
         }
         op1.setCaseSensitive(true);
         if (CompareOps.equals(op1, op2) != cEq)
         {
            flag = false;
            System.out.println("isEqual(c) failure for \"" + op1 +
                                "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(false);
         if (CompareOps.notEqual(op1, op2) != ne)
         {
            flag = false;
            System.out.println("notEqual failure for \"" + op1 + "\" and \"" +
                               op2 + "\"");
         }
         op1.setCaseSensitive(true);
         if (CompareOps.notEqual(op1, op2) != cNe)
         {
            flag = false;
            System.out.println("notEqual(c) failure for \"" + op1 +
                                "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(false);
         if (CompareOps.greaterThan(op1, op2) != gt)
         {
            flag = false;
            System.out.println("isGreaterThan failure for \"" + op1 +
                               "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(true);
         if (CompareOps.greaterThan(op1, op2) != cGt)
         {
            flag = false;
            System.out.println("isGreaterThan(c) failure for \"" + op1 +
                                "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(false);
         if (CompareOps.lessThan(op1, op2) != lt)
         {
            flag = false;
            System.out.println("isLessThan failure for \"" + op1 +
                               "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(true);
         if (CompareOps.lessThan(op1, op2) != cLt)
         {
            flag = false;
            System.out.println("isLessThan(c) failure for \"" + op1 +
                                "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(false);
         if (CompareOps.greaterThanOrEqual(op1, op2) != ge)
         {
            flag = false;
            System.out.println("isGreaterThanOrEqual failure for \"" + op1 +
                               "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(true);
         if (CompareOps.greaterThanOrEqual(op1, op2) != cGe)
         {
            flag = false;
            System.out.println("isGreaterThanOrEqual(c) failure for \"" +
                                op1 + "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(false);
         if (CompareOps.lessThanOrEqual(op1, op2) != le)
         {
            flag = false;
            System.out.println("isLessThanOrEqual failure for \"" + op1 +
                               "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(true);
         if (CompareOps.lessThanOrEqual(op1, op2) != cLe)
         {
            flag = false;
            System.out.println("isLessThanOrEqual(c) failure for \"" + op1 +
                                "\" and \"" + op2 + "\"");
         }
         op1.setCaseSensitive(false);
         
         if (TextOps._matches(op1, op2) != matches)
         {
            flag = false;
            System.out.println("matches failure for \"" + this.op1 + 
                               "\" (\"" + op1 + "\") and \"" + this.op2 + 
                               "\" (\"" + op2 + "\")");
         }
         op1.setCaseSensitive(true);
         if (TextOps._matches(op1, op2) != cMatches)
         {
            flag = false;
            System.out.println("matches(c) failure for \"" + this.op1 +
                                "\" (\"" + op1 + "\") and \"" + this.op2 +
                                "\" (\"" + op2 + "\")");
         }
         
         return flag;
      }
   }
   
   /**
    * Command line interface which provides a test harness to regression
    * test this class.  No parameters are necessary.
    */
   public static void main(String[] args)
   {            
      // if still true after all tests, everything passed
      boolean flag = true;
           
      TestCase tests[] =
      {
         new TestCase(" hello "    , "hello"      , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase("hello"      , "hello  "    , false, true , false, false,
                                                    false, true , true , false,
                                                    false, true , false, false,
                                                    false, true , true , false),
         new TestCase("hello    "  , "hello"      , true , true , false, false,
                                                    false, true , true , false,
                                                    true , true , false, false,
                                                    false, true , true , false),
         new TestCase("hELLo "     , "hello"      , true , true , false, false,
                                                    false, true , true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase("hello    "  , "heLLo"      , true , true , false, false,
                                                    false, true , true , false,
                                                    false, false, true , true ,
                                                    false, true , false, false),
         new TestCase(" hello "    , ".ell."      , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase(" hello "    , ".*ell*"     , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase(" hello~hel", "..ell.\\*"   , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase(" hel~lo "    , " hell*"    , false, false, true , true ,
                                                    false, true , false, true ,
                                                    false, false, true , true ,
                                                    false, true , false, true ),
         new TestCase(" ajaj\\aj " , "*aj*"       , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase(" ajaj\\aj " , "*aj\\aj "   , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase(" ajaj\\aj " , "*ajaj "     , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase(" ajaj\\\\aj " , "*ajaj "   , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase(" ajaj\\\\\\aj " , "*ajaj " , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase(" ajaj\\\\aj " , "*aj\\aj " , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase(" ajaj\\\\aj " , "*aj\\\\aj ", false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase(" ajaj\\\\aj ", " ajaj\\\\aj ", true , true , false, false,
                                                    false, true , true , false,
                                                    true , true , false, false,
                                                    false, true , true , false),
         new TestCase(" ajaj~aj " , "*aj*"        , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase(" ajaj~aj "   , "*aj~~aj "  , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase(" ajaj~aj "   , "*ajaj "    , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase(" ajaj~~aj "  , "*ajaj "    , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase(" ajaj~~~aj "  , "*ajaj "   , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase(" ajaj~~aj "  , "*aj~aj "   , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase(" ajaj~~aj "  , "*aj~~aj "  , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase(" ajaj~~aj "  , " ajaj~~aj ", true , true , false, false,
                                                    false, true , true , false,
                                                    true , true , false, false,
                                                    false, true , true , false),
         new TestCase("a\\a"         , "a~~a"     , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase("a\\a"         , "a\\\\a"   , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , true ,
                                                    false, true , false, true ),
         new TestCase("a\\\\a"       , "a\\\\\\\\a", false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , true ,
                                                    false, true , false, true ),
         new TestCase("a~~a"        , "a~~~~a"    , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase("*a.b"       , "~~*a~~.b"     , false, false, true , false,
                                                    true , false, true , true ,
                                                    false, false, true , false,
                                                    true , false, true , true ),
         new TestCase("*a.b"       , "~*a~.b"       , true , true , false, false,
                                                    false, true , true , true ,
                                                    true , true , false, false,
                                                    false, true , true , true ),
         new TestCase("*a.b"       , "*a.b"       , true , true , false, false,
                                                    false, true , true , true ,
                                                    true , true , false, false,
                                                    false, true , true , true ),
         new TestCase("*axb"       , "~~*a~~.b"     , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase("*axb"       , "~*a~.b"       , false, false, true , true ,
                                                    false, true , false, true ,
                                                    false, false, true , true ,
                                                    false, true , false, true ),
         new TestCase("*axb"       , "*a.b"       , false, false, true , true ,
                                                    false, true , false, true ,
                                                    false, false, true , true ,
                                                    false, true , false, true ),
         new TestCase("xa.b"       , "~~*a~~.b"     , false, false, true , false,
                                                    true , false, true , false,
                                                    false, false, true , false,
                                                    true , false, true , false),
         new TestCase("xa.b"       , "~*a~.b"       , false, false, true , true ,
                                                    false, true , false, true ,
                                                    false, false, true , true ,
                                                    false, true , false, true ),
         new TestCase("xa.b"       , "*a.b"       , false, false, true , true ,
                                                    false, true , false, true ,
                                                    false, false, true , true ,
                                                    false, true , false, true )
      };      
          
      // run the tests
      for (int i = 0; i < tests.length; i++)
      {
         if (tests[i].test() == false)
            flag = false;
      }
      
      if (flag)
      {
         System.out.println("All tests passed successfully.");
      }

      System.out.println(TextOps.matchesList("abc,def", "ABc") + " yes");
      System.out.println(TextOps.matchesList("abc,def", "dEF") + " yes");
      System.out.println(TextOps.matchesList("abc,def", "xyz") + " no ");
      System.out.println(TextOps.matchesList("!abc,def", "ABc") + " no");
      System.out.println(TextOps.matchesList("!abc,def", "dEF") + " yes");
      System.out.println(TextOps.matchesList("abc*,def", "ABc") + " yes");
      System.out.println(TextOps.matchesList("ab*,def", "ABc")  + " yes");
      System.out.println(TextOps.matchesList("a*b*c,def", "AxBc") + " yes");
      System.out.println(TextOps.matchesList("a*b*c,def", "AxByc") + " yes");
      System.out.println(TextOps.matchesList("!a*b*c,def", "AxBc") + " no ");
      System.out.println(TextOps.matchesList("!a*b*c,def", "AxByc") + " no ");
      System.out.println(TextOps.matchesList("a!bc,def", "A!Bc") + " yes");
      System.out.println(TextOps.matchesList("abc,d!ef", "d!EF") + " yes");
      System.out.println(TextOps.matchesList(",abc,def", "") + " yes");
      System.out.println(TextOps.matchesList(" ,abc,def", "") + " no ");
      System.out.println(TextOps.matchesList(" ,abc,def", " ") + " yes");
      System.out.println(TextOps.matchesList("abc,def,", "") + " yes");
      System.out.println(TextOps.matchesList("abc,def,", " ") + " no ");
      System.out.println(TextOps.matchesList("abc,def, ", " ") + " yes");
      System.out.println(TextOps.matchesList("", " ") + " no ");
      System.out.println(TextOps.matchesList("", "") + " yes");
      System.out.println(TextOps.matchesList(new character(), " ") + " ?  ");
      System.out.println(TextOps.matchesList(" ", new character()) + " ?  ");
      System.out.println(TextOps.matchesList(new character(), new character()) + " ?  ");
      System.out.println(TextOps.matchesList(new character()) + " exception");

   }
   
   // This is the Progress 4GL program that was used to generate the test
   // case results.  As output it generates the Java statements to initialize
   // the tests[] array in main(), HOWEVER due to the natural conversion of
   // string literals by Progress (on input), the corresponding output
   // strings are missing tildes and backslashes... so you will have to fix
   // up these strings to include the missing data.  In addition, since Java
   // modifies string literals at compile time, you will need to double up
   // every backslash to ensure that the resulting number of backslashes
   // match what the Progress program used.  So if there were 2 backslashes
   // in the Progress version of a string, the Java version will need 4.
   // Once these fixups occur in the main() method, then the tests can be
   // run to determine the level of compatibility between this class and
   // Progress string comparison operations.
   /*   
   
   def var op1  as char extent 38.
   def var op1_ as char extent 38 case-sensitive.
   def var op2  as char extent 38.
   def var res  as logical format "true/false".
   def var i    as int.
   def var str  as char.
   
   op1[1]  = " hello ".     op2[1]  = "hello".
   op1[2]  = "hello".       op2[2]  = "hello  ".
   op1[3]  = "hello    ".   op2[3]  = "hello".    
   op1[4]  = "hELLo ".      op2[4]  = "hello".
   op1[5]  = "hello    ".   op2[5]  = "heLLo".
   op1[6]  = " hello ".     op2[6]  = ".ell.".
   op1[7]  = " hello ".     op2[7]  = ".*ell*".
   op1[8]  = " hello~hel".  op2[8]  = "..ell.\*".
   op1[9]  = " hel~lo ".    op2[9]  = " hell*".
   op1[10] = " ajaj\aj ".   op2[10] = "*aj*".
   op1[11] = " ajaj\aj ".   op2[11] =  "*aj\aj ".
   op1[12] = " ajaj\aj ".   op2[12] =  "*ajaj ".
   op1[13] = " ajaj\\aj ".  op2[13] =  "*ajaj ".
   op1[14] = " ajaj\\\aj ". op2[14] =  "*ajaj ".
   op1[15] = " ajaj\\aj ".  op2[15] =  "*aj\aj ".
   op1[16] = " ajaj\\aj ".  op2[16] =  "*aj\\aj ".
   op1[17] = " ajaj\\aj ".  op2[17] =  " ajaj\\aj ".
   op1[18] = " ajaj~aj ".   op2[18] =  "*aj*".
   op1[19] = " ajaj~aj ".   op2[19] =  "*aj~aj ".
   op1[20] = " ajaj~aj ".   op2[20] =  "*ajaj ".
   op1[21] = " ajaj~~aj ".  op2[21] =  "*ajaj ".
   op1[22] = " ajaj~~~aj ". op2[22] =  "*ajaj ".
   op1[23] = " ajaj~~aj ".  op2[23] =  "*aj~aj ".
   op1[24] = " ajaj~~aj ".  op2[24] =  "*aj~~aj ".
   op1[25] = " ajaj~~aj ".  op2[25] =  " ajaj~~aj ".
   op1[26] = "a\a".         op2[26] =  "a~~a".
   op1[27] = "a\a".         op2[27] =  "a\\a".
   op1[28] = "a\\a".        op2[28] =  "a\\\\a".
   op1[29] = "a~~a".        op2[29] =  "a~~~~a".
   op1[30] = "*a.b".        op2[30] =  "~~*a~~.b".
   op1[31] = "*a.b".        op2[31] =  "~*a~.b".
   op1[32] = "*a.b".        op2[32] =  "*a.b".
   op1[33] = "*axb".        op2[33] =  "~~*a~~.b".
   op1[34] = "*axb".        op2[34] =  "~*a~.b".
   op1[35] = "*axb".        op2[35] =  "*a.b".
   op1[36] = "xa.b".        op2[36] =  "~~*a~~.b".
   op1[37] = "xa.b".        op2[37] =  "~*a~.b".
   op1[38] = "xa.b".        op2[38] =  "*a.b".
   
   output to value("listing.txt").
   
   do i = 1 to 38.
      str = "         new TestCase(\"" + op1[i] + "\"" +
            fill(" ", 11 - length(op1[i])) + ", \"" + op2[i] + "\"" +
            fill(" ", 11 - length(op2[i])) + ", ". 
      put str format "x(52)".
   
      res = op1[i] begins op2[i].
      put res ", ".
      res = op1[i] eq op2[i].
      put res ", ".   
      res = op1[i] ne op2[i].
      put res ", ".
      res = op1[i] gt op2[i].
      put res ",~n".
      put fill(" ", 52) format "x(52)".
      res = op1[i] lt op2[i].
      put res ", ".   
      res = op1[i] ge op2[i].
      put res ", ".   
      res = op1[i] le op2[i].
      put res ", ".   
      res = op1[i] matches op2[i].
      put res ",~n".   
      put fill(" ", 52) format "x(52)".
      
      op1_[i] = op1[i].
      
      res = op1_[i] begins op2[i].
      put res ", ".
      res = op1_[i] eq op2[i].
      put res ", ".   
      res = op1_[i] ne op2[i].
      put res ", ".
      res = op1_[i] gt op2[i].
      put res ",~n".
      put fill(" ", 52) format "x(52)".
      res = op1_[i] lt op2[i].
      put res ", ".   
      res = op1_[i] ge op2[i].
      put res ", ".   
      res = op1_[i] le op2[i].
      put res ", ".   
      res = op1_[i] matches op2[i].
      put res "),~n".      
   end.
   
   output close.
   
   */   
}
